#! /usr/bin/perl

################################################################################
# Script to update a Perforce server's changelists using the change map file
# generated by p4convert-svn.  Specifically, this updates a given changelist's
# Date and User fields.
#
# Requires P4Perl and a user account on your Perforce server that has at least
# "Admin" level access.
################################################################################

use strict;
use Getopt::Long;
use DateTime;        
use DateTime::TimeZone;

use P4;
my $p4 = new P4;

my $p4port     = "";
my $clientname = "";
my $p4password = "";
my $userprefix = "";
my $usersuffix = "";
my $startRev = 1;
my $stopRev = -1;
my $non_interactive = 0;
my $localTimeZone;

GetOptions ('port=s' => \$p4port, 'client=s' => \$clientname,
            'password=s' => \$p4password, 'userprefix=s' => \$userprefix,
            'usersuffix=s' => \$usersuffix, 'startrev=i' => \$startRev,
            'stoprev=i' => \$stopRev, 'non-interactive' => \$non_interactive,
            'localtimezone=s' => \$localTimeZone);

################################################################################
# Connect to the server.
$p4->SetClient( $clientname );
$p4->SetPort( $p4port );
$p4->SetPassword( $p4password );
$p4->Connect()
    or die( "Failed to connect to Perforce Server.  Verify P4CONFIG setup.\n" );

# Date conversion won't work without a time zone.
die "A local time zone must be defined!" unless defined( $localTimeZone );

# Can't go past the last change in the system
my $lastRev = $p4->RunCounter('change')->[0]->{ 'value' };
if( $stopRev > $lastRev || $stopRev == -1 )
{
    $stopRev = $lastRev;
}

if( $startRev > $stopRev )
{
    $startRev = $stopRev;
}

# Keep track of valid users.
my @users = $p4->RunUsers();
my %user_present;
foreach my $user ( @users )
{
    $user_present{ $user->{ 'User' } } = 1;
}

print "\nThis is the \"p4 info\" for the server you're about to use:\n\n";

my @info = $p4->RunInfo();
foreach ( @info )
{
    foreach my $k ( keys %$_ )
    {
        my $v = $$_{ $k };
        print"$k:  $v\n";
    }
}

print "\nOptions for this update:\n";
print "Start Revision: $startRev\n";
print "Stop Revision: $stopRev\n";
print "User Prefix: $userprefix\n";
print "User Suffix: $usersuffix\n\n";
print "Local Timezone: $localTimeZone\n\n";

print "\n\nAbout to update the user name, date and change description of\n";
print "existing, submitted changelists starting with revision $startRev.\n";
print "Only do this if you're certain you have the correct server and data,\n";
print "as it might be difficult to reverse if it's done in error.\n";
print "\nProceed?  (yes/no)  ";
 
if( $non_interactive == 0 )
{
    my $prompt = <STDIN>;
    chomp( $prompt );
    if( $prompt eq "no" || $prompt ne "yes" )
    {
        print "\nAborting changelist update.\n";
        exit;
    }
}
print "\n";

################################################################################
# Parse the selected range of changesets and update any that
# have not already been updated.

my $count = 0;
my $change;
for ( $change = $startRev; $change <= $stopRev; $change++ )
{
    my $changeSpec = $p4->FetchChange( $change );

    # move svn data to end of changelist description
    my @description = split /\n/, $changeSpec->{'Description'};
    my $description_text = "";
    my $conversion_text = "";
    my $user;
    my $date;
    foreach my $line ( @description )
    {
        # svn rev 2 Author:rgrimes Date:1993-06-12T14:49:13.000000Z
        #  or
        # PARTIAL(rollback) svn rev 203843 Author:marius Date:2010-02-13T18:18:45.622967Z
        if( $line =~ m/^\s*(PARTIAL\([^)]+\)\s+)?svn rev \d+ Author:(\S+) Date:(\d+)-(\d+)-(\d+)T(\d+):(\d+):(\d+)\.\d+[A-Z]/ )
        {

            $user = $2;
            my $utcTime = DateTime->new(year => $3, month => $4, day => $5,
                                        hour => $6, minute => $7, second => $8,
                                        time_zone => 'UTC');
            my $localTime = $utcTime->clone()->set_time_zone( $localTimeZone );
            $date = $localTime->strftime("%Y/%m/%d %H:%M:%S");
            $conversion_text = "SVN Importer: " .  $conversion_text
                             . $line . "\n";
        }
        else
        {
            $description_text =  $description_text . $line . "\n";
        }
    }

    # Don't update already converted changes.
    next if( $conversion_text eq "" );

    my $changeUser = $userprefix . $user . $usersuffix;
    ensure_user( $changeUser );

    print "Updating changelist $change.\n";
    $changeSpec->{'Description'} = $description_text . "\n" . $conversion_text;
    $changeSpec->{'User'} = $changeUser;
    $changeSpec->{'Date'} = $date;
    $p4->SetInput( $changeSpec );
    $p4->Run( "change", "-f", "-i" );
    die_on_p4error( "Change submission failed." );

    $count++;
}

################################################################################
# Conversion complete.
cleanup();
exit;

my $tempUser = "";

################################ Subroutines ###################################
# Create a temporary user to appease 2009.2 which requires
# the user to exist for the username to be changed via 
# 'change -f'
sub ensure_user
{
    my $user = shift;
    if( $tempUser ne $user && not $user_present{ $user } )
    {
        # Free up any existing temporary user.  We don't
        # want to use more than one license slot at a time.
        cleanup_user();
        my $userSpec = $p4->FetchUser( $user );
        $p4->SetInput( $userSpec );
        $p4->Run( "user", "-f", "-i" );
        die_on_p4error( "Unable to create user $user!" );
        $tempUser = $user;
    }
}

sub cleanup_user
{
    if( $tempUser ne "" )
    {
        # Delete temporary user
        $p4->DeleteUser( "-f", $tempUser );
        die_on_p4error( "Unable to delete user $tempUser!" );
        $tempUser = "";
    }

}

sub die_on_p4error
{
    my $error_string = shift;

    if( $p4->ErrorCount() > 0 )
    {
        # Record that we've stopped early.
        $stopRev = $change;
        cleanup();
        die $error_string . "\n\n" . $p4->Errors();
    }
}

sub cleanup
{
    cleanup_user();
    $p4->Disconnect();
    print "\n\nProcessed revisions $startRev - $stopRev, $count updated\n";
}
